def mermaid_assert(output: str) -> None:
    assert "users {" in output
    assert "posts {" in output
    assert "comments {" in output

    assert "users ||--o{ posts : author" in output
    assert "posts ||--o{ comments : post" in output
    assert "users ||--o{ comments : author" in output

    assert "CHAR(32) author FK" in output
    assert 'CHAR(32) post FK "nullable"' in output
    assert 'BOOLEAN live "True if post is published,nullable"' in output
    assert "DATETIME created" in output



from paracelsus.pyproject import get_pyproject_settings


def test_pyproject(package_path):
    settings = get_pyproject_settings(package_path)
    assert "base" in settings
    assert "imports" in settings
    assert settings["base"] == "example.base:Base"
    assert "example.models" in settings["imports"]


def test_pyproject_none():
    settings = get_pyproject_settings()
    assert settings is None



import os
import shutil
import tempfile
from datetime import datetime, timezone
from pathlib import Path
from uuid import uuid4

import pytest
from sqlalchemy import Boolean, DateTime, ForeignKey, String, Text, Uuid
from sqlalchemy.orm import declarative_base, mapped_column

UTC = timezone.utc


@pytest.fixture
def metaclass():
    Base = declarative_base()

    class User(Base):
        __tablename__ = "users"

        id = mapped_column(Uuid, primary_key=True, default=uuid4())
        display_name = mapped_column(String(100))
        created = mapped_column(DateTime, nullable=False, default=datetime.now(UTC))

    class Post(Base):
        __tablename__ = "posts"

        id = mapped_column(Uuid, primary_key=True, default=uuid4())
        author = mapped_column(ForeignKey(User.id), nullable=False)
        created = mapped_column(DateTime, nullable=False, default=datetime.now(UTC))
        live = mapped_column(Boolean, default=False, comment="True if post is published")
        content = mapped_column(Text, default="")

    class Comment(Base):
        __tablename__ = "comments"

        id = mapped_column(Uuid, primary_key=True, default=uuid4())
        post = mapped_column(Uuid, ForeignKey(Post.id), default=uuid4())
        author = mapped_column(ForeignKey(User.id), nullable=False)
        created = mapped_column(DateTime, nullable=False, default=datetime.now(UTC))
        live = mapped_column(Boolean, default=False)
        content = mapped_column(Text, default="")

    return Base.metadata


@pytest.fixture
def package_path():
    template_path = Path(os.path.dirname(os.path.realpath(__file__))) / "assets"
    with tempfile.TemporaryDirectory() as package_path:
        shutil.copytree(template_path, package_path, dirs_exist_ok=True)
        yield Path(package_path)



from paracelsus.graph import get_graph_string

from .utils import mermaid_assert


def test_get_graph_string(package_path):
    graph_string = get_graph_string(
        base_class_path="example.base:Base",
        import_module=["example.models"],
        include_tables=set(),
        exclude_tables=set(),
        python_dir=[package_path],
        format="mermaid",
    )
    mermaid_assert(graph_string)


def test_get_graph_string_with_exclude(package_path):
    """Excluding tables removes them from the graph string."""
    graph_string = get_graph_string(
        base_class_path="example.base:Base",
        import_module=["example.models"],
        include_tables=set(),
        exclude_tables={"comments"},
        python_dir=[package_path],
        format="mermaid",
    )
    assert "comments {" not in graph_string
    assert "posts {" in graph_string
    assert "users {" in graph_string
    assert "users ||--o{ posts" in graph_string

    # Excluding a table to which another table holds a foreign key will raise an error.
    graph_string = get_graph_string(
        base_class_path="example.base:Base",
        import_module=["example.models"],
        include_tables=set(),
        exclude_tables={"users", "comments"},
        python_dir=[package_path],
        format="mermaid",
    )
    assert "posts {" in graph_string
    assert "users ||--o{ posts" not in graph_string


def test_get_graph_string_with_include(package_path):
    """Excluding tables keeps them in the graph string."""
    graph_string = get_graph_string(
        base_class_path="example.base:Base",
        import_module=["example.models"],
        include_tables={"users", "posts"},
        exclude_tables=set(),
        python_dir=[package_path],
        format="mermaid",
    )
    assert "comments {" not in graph_string
    assert "posts {" in graph_string
    assert "users {" in graph_string
    assert "users ||--o{ posts" in graph_string

    # Including a table that holds a foreign key to a non-existing table will keep
    # the table but skip the connection.
    graph_string = get_graph_string(
        base_class_path="example.base:Base",
        import_module=["example.models"],
        include_tables={"posts"},
        exclude_tables=set(),
        python_dir=[package_path],
        format="mermaid",
    )
    assert "posts {" in graph_string
    assert "users ||--o{ posts" not in graph_string



from sqlalchemy.sql.schema import Column


def column_sort_key(column: Column):
    if column.primary_key:
        prefix = "01"
    elif len(column.foreign_keys):
        prefix = "02"
    else:
        prefix = "03"
    return f"{prefix}_{column.name}"



import logging
from sqlalchemy.sql.schema import Column, MetaData, Table

from . import utils


logger = logging.getLogger(__name__)


class Mermaid:
    comment_format: str = "mermaid"
    metadata: MetaData

    def __init__(self, metaclass: MetaData) -> None:
        self.metadata = metaclass

    def _table(self, table: Table) -> str:
        output = f"  {table.name}"
        output += " {\n"
        columns = sorted(table.columns, key=utils.column_sort_key)
        for column in columns:
            output += self._column(column)
        output += "  }\n\n"
        return output

    def _column(self, column: Column) -> str:
        options = []
        column_str = f"{column.type} {column.name}"

        if column.primary_key:
            if len(column.foreign_keys) > 0:
                column_str += " PK,FK"
            else:
                column_str += " PK"
        elif len(column.foreign_keys) > 0:
            column_str += " FK"
        elif column.unique:
            column_str += " UK"

        if column.comment:
            options.append(column.comment)

        if column.nullable:
            options.append("nullable")

        if column.index:
            options.append("indexed")

        if len(options) > 0:
            column_str += f' "{",".join(options)}"'

        return f"    {column_str}\n"

    def _relationships(self, column: Column) -> str:
        output = ""

        column_name = column.name
        right_table = column.table.name

        if column.unique:
            right_operand = "o|"
        else:
            right_operand = "o{"

        for foreign_key in column.foreign_keys:
            key_parts = foreign_key.target_fullname.split(".")
            left_table = key_parts[0]
            left_column = key_parts[1]
            left_operand = ""

            # We don't add the connection to the fk table if the latter
            # is not included in our graph.
            if left_table not in self.metadata.tables:
                logger.warning(
                    f"Table '{right_table}.{column_name}' is a foreign key to '{left_table}' "
                    "which is not included in the graph, skipping the connection."
                )
                continue

            lcolumn = self.metadata.tables[left_table].columns[left_column]
            if lcolumn.unique or lcolumn.primary_key:
                left_operand = "||"
            else:
                left_operand = "}o"

            output += f"  {left_table} {left_operand}--{right_operand} {right_table} : {column_name}\n"
        return output

    def __str__(self) -> str:
        output = "erDiagram\n"
        for table in self.metadata.tables.values():
            output += self._table(table)

        for table in self.metadata.tables.values():
            for column in table.columns.values():
                if len(column.foreign_keys) > 0:
                    output += self._relationships(column)

        return output



import importlib
import os
import sys
from pathlib import Path
from typing import List, Set

from sqlalchemy import MetaData

from .transformers.dot import Dot
from .transformers.mermaid import Mermaid

transformers = {
    "mmd": Mermaid,
    "mermaid": Mermaid,
    "dot": Dot,
    "gv": Dot,
}


def get_graph_string(
    *,
    base_class_path: str,
    import_module: List[str],
    include_tables: Set[str],
    exclude_tables: Set[str],
    python_dir: List[Path],
    format: str,
) -> str:
    # Update the PYTHON_PATH to allow more module imports.
    sys.path.append(str(os.getcwd()))
    for dir in python_dir:
        sys.path.append(str(dir))

    # Import the base class so the metadata class can be extracted from it.
    # The metadata class is passed to the transformer.
    module_path, class_name = base_class_path.split(":", 2)
    base_module = importlib.import_module(module_path)
    base_class = getattr(base_module, class_name)
    metadata = base_class.metadata

    # The modules holding the model classes have to be imported to get put in the metaclass model registry.
    # These modules aren't actually used in any way, so they are discarded.
    # They are also imported in scope of this function to prevent namespace pollution.
    for module in import_module:
        if ":*" in module:
            # Sure, execs are gross, but this is the only way to dynamically import wildcards.
            exec(f"from {module[:-2]} import *")
        else:
            importlib.import_module(module)

    # Grab a transformer.
    if format not in transformers:
        raise ValueError(f"Unknown Format: {format}")
    transformer = transformers[format]

    # Keep only the tables which were included / not-excluded
    include_tables = resolve_included_tables(
        include_tables=include_tables, exclude_tables=exclude_tables, all_tables=set(metadata.tables.keys())
    )
    filtered_metadata = filter_metadata(metadata=metadata, include_tables=include_tables)

    # Save the graph structure to string.
    return str(transformer(filtered_metadata))


def resolve_included_tables(
    include_tables: Set[str],
    exclude_tables: Set[str],
    all_tables: Set[str],
) -> Set[str]:
    """Resolves the final set of tables to include in the graph.

    Given sets of inclusions and exclusions and the set of all tables we define
    the following cases are:
    - Empty inclusion and empty exclusion -> include all tables.
    - Empty inclusion and some exclusions -> include all tables except the ones in the exclusion set.
    - Some inclusions and empty exclusion -> make sure tables in the inclusion set are present in
        all tables then include the tables in the inclusion set.
    - Some inclusions and some exclusions -> not resolvable, an error is raised.
    """
    match len(include_tables), len(exclude_tables):
        case 0, 0:
            return all_tables
        case 0, int():
            return all_tables - exclude_tables
        case int(), 0:
            if not include_tables.issubset(all_tables):
                non_existent_tables = include_tables - all_tables
                raise ValueError(
                    f"Some tables to include ({non_existent_tables}) don't exist"
                    "withinthe found tables ({all_tables})."
                )
            return include_tables
        case _:
            raise ValueError(
                f"Only one or none of include_tables ({include_tables}) or exclude_tables"
                f"({exclude_tables}) can contain values."
            )


def filter_metadata(
    metadata: MetaData,
    include_tables: Set[str],
) -> MetaData:
    """Create a subset of the metadata based on the tables to include."""
    filtered_metadata = MetaData()
    for tablename, table in metadata.tables.items():
        if tablename in include_tables:
            table.tometadata(filtered_metadata)
    return filtered_metadata



import os
from pathlib import Path
from typing import Any, Dict

try:
    import tomllib
except:  # noqa: E722
    import toml as tomllib  # type: ignore


def get_pyproject_settings(dir: Path = Path(os.getcwd())) -> Dict[str, Any] | None:
    pyproject = dir / "pyproject.toml"

    if not pyproject.exists():
        return None

    with open(pyproject, "rb") as f:
        data = tomllib.loads(f.read().decode())

    return data.get("tool", {}).get("paracelsus", None)



from paracelsus.transformers.mermaid import Mermaid

from ..utils import mermaid_assert


def test_mermaid(metaclass):
    mermaid = Mermaid(metaclass=metaclass)
    graph_string = str(mermaid)
    mermaid_assert(graph_string)



import pydot  # type: ignore
import logging
from sqlalchemy.sql.schema import MetaData, Table

from . import utils


logger = logging.getLogger(__name__)


class Dot:
    comment_format: str = "dot"
    metadata: MetaData
    graph: pydot.Dot

    def __init__(self, metaclass: MetaData) -> None:
        self.metadata = metaclass
        self.graph = pydot.Dot("database", graph_type="graph")

        for table in self.metadata.tables.values():
            node = pydot.Node(name=table.name)
            node.set_label(self._table_label(table))
            node.set_shape("none")
            node.set_margin("0")
            self.graph.add_node(node)
            for column in table.columns:
                for foreign_key in column.foreign_keys:
                    key_parts = foreign_key.target_fullname.split(".")
                    left_table = key_parts[0]
                    left_column = key_parts[1]

                    # We don't add the connection to the fk table if the latter
                    # is not included in our graph.
                    if left_table not in self.metadata.tables:
                        logger.warning(
                            f"Table '{table}.{column.name}' is a foreign key to '{left_table}' "
                            "which is not included in the graph, skipping the connection."
                        )
                        continue

                    edge = pydot.Edge(left_table, table.name)
                    edge.set_label(column.name)
                    edge.set_dir("both")

                    edge.set_arrowhead("none")
                    if not column.unique:
                        edge.set_arrowhead("crow")

                    l_column = self.metadata.tables[left_table].columns[left_column]
                    edge.set_arrowtail("none")
                    if not l_column.unique and not l_column.primary_key:
                        edge.set_arrowtail("crow")

                    self.graph.add_edge(edge)

    def _table_label(self, table: Table) -> str:
        column_output = ""
        columns = sorted(table.columns, key=utils.column_sort_key)
        for column in columns:
            attributes = set([])
            if column.primary_key:
                attributes.add("Primary Key")

            if len(column.foreign_keys) > 0:
                attributes.add("Foreign Key")

            if column.unique:
                attributes.add("Unique")

            column_output += f'        <tr><td align="left">{column.type}</td><td align="left">{column.name}</td><td>{", ".join(sorted(attributes))}</td></tr>\n'

        return f"""<
    <table border="0" cellborder="1" cellspacing="0" cellpadding="4">
        <tr><td colspan="3" bgcolor="lightblue"><b>{table.name}</b></td></tr>
{column_output.rstrip()}
    </table>
>"""

    def __str__(self) -> str:
        return self.graph.to_string()



from typer.testing import CliRunner

from paracelsus.cli import app

from .utils import mermaid_assert

runner = CliRunner()


def test_graph(package_path):
    result = runner.invoke(
        app,
        [
            "graph",
            "example.base:Base",
            "--import-module",
            "example.models",
            "--python-dir",
            str(package_path),
        ],
    )

    assert result.exit_code == 0
    mermaid_assert(result.stdout)


def test_graph_with_exclusion(package_path):
    result = runner.invoke(
        app,
        [
            "graph",
            "example.base:Base",
            "--import-module",
            "example.models",
            "--python-dir",
            str(package_path),
            "--exclude-tables",
            "comments",
        ],
    )
    assert result.exit_code == 0
    assert "posts {" in result.stdout
    assert "comments {" not in result.stdout


def test_graph_with_inclusion(package_path):
    result = runner.invoke(
        app,
        [
            "graph",
            "example.base:Base",
            "--import-module",
            "example.models",
            "--python-dir",
            str(package_path),
            "--include-tables",
            "comments",
        ],
    )
    assert result.exit_code == 0
    assert "posts {" not in result.stdout
    assert "comments {" in result.stdout


def test_inject_check(package_path):
    result = runner.invoke(
        app,
        [
            "inject",
            str(package_path / "README.md"),
            "example.base:Base",
            "--import-module",
            "example.models",
            "--python-dir",
            str(package_path),
            "--check",
        ],
    )
    assert result.exit_code == 1


def test_inject(package_path):
    result = runner.invoke(
        app,
        [
            "inject",
            str(package_path / "README.md"),
            "example.base:Base",
            "--import-module",
            "example.models",
            "--python-dir",
            str(package_path),
        ],
    )
    assert result.exit_code == 0

    with open(package_path / "README.md") as fp:
        readme = fp.read()
        mermaid_assert(readme)


def test_version():
    result = runner.invoke(app, ["version"])
    assert result.exit_code == 0



from paracelsus.transformers.dot import Dot


def test_mermaid(metaclass):
    mermaid = Dot(metaclass=metaclass)
    graph_string = str(mermaid)

    assert '<tr><td colspan="3" bgcolor="lightblue"><b>users</b></td></tr>' in graph_string
    assert '<tr><td colspan="3" bgcolor="lightblue"><b>posts</b></td></tr>' in graph_string
    assert '<tr><td colspan="3" bgcolor="lightblue"><b>comments</b></td></tr>' in graph_string

    assert "users -- posts  [arrowhead=crow, arrowtail=none, dir=both, label=author];" in graph_string
    assert "posts -- comments  [arrowhead=crow, arrowtail=none, dir=both, label=post];" in graph_string
    assert "users -- comments  [arrowhead=crow, arrowtail=none, dir=both, label=author];" in graph_string

    assert '<tr><td align="left">CHAR(32)</td><td align="left">author</td><td>Foreign Key</td></tr>' in graph_string
    assert '<tr><td align="left">CHAR(32)</td><td align="left">post</td><td>Foreign Key</td></tr>' in graph_string
    assert '<tr><td align="left">DATETIME</td><td align="left">created</td><td></td></tr>' in graph_string



from datetime import datetime, timezone
from uuid import uuid4

from sqlalchemy import Boolean, DateTime, ForeignKey, String, Text, Uuid
from sqlalchemy.orm import mapped_column

from .base import Base

UTC = timezone.utc


class User(Base):
    __tablename__ = "users"

    id = mapped_column(Uuid, primary_key=True, default=uuid4())
    display_name = mapped_column(String(100))
    created = mapped_column(DateTime, nullable=False, default=datetime.now(UTC))


class Post(Base):
    __tablename__ = "posts"

    id = mapped_column(Uuid, primary_key=True, default=uuid4())
    author = mapped_column(ForeignKey(User.id), nullable=False)
    created = mapped_column(DateTime, nullable=False, default=datetime.now(UTC))
    live = mapped_column(Boolean, default=False, comment="True if post is published")
    content = mapped_column(Text, default="")


class Comment(Base):
    __tablename__ = "comments"

    id = mapped_column(Uuid, primary_key=True, default=uuid4())
    post = mapped_column(Uuid, ForeignKey(Post.id), default=uuid4())
    author = mapped_column(ForeignKey(User.id), nullable=False)
    created = mapped_column(DateTime, nullable=False, default=datetime.now(UTC))
    live = mapped_column(Boolean, default=False)
    content = mapped_column(Text, default="")
    content = mapped_column(Text, default="")



import re
import sys
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional

import typer
from typing_extensions import Annotated

from .graph import get_graph_string, transformers
from .pyproject import get_pyproject_settings

app = typer.Typer()


class Formats(str, Enum):
    mermaid = "mermaid"
    mmd = "mmd"
    dot = "dot"
    gv = "gv"


def get_base_class(base_class_path: str | None, settings=Dict[str, Any] | None) -> str:
    if base_class_path:
        return base_class_path
    if not settings:
        raise ValueError("`base_class_path` argument must be passed if no pyproject.toml file is present.")
    if "base" not in settings:
        raise ValueError("`base_class_path` argument must be passed if not defined in pyproject.toml.")
    return settings["base"]


@app.command(help="Create the graph structure and print it to stdout.")
def graph(
    base_class_path: Annotated[
        Optional[str],
        typer.Argument(help="The SQLAlchemy base class used by the database to graph."),
    ] = None,
    import_module: Annotated[
        List[str],
        typer.Option(
            help="Module, typically an SQL Model, to import. Modules that end in :* will act as `from module import *`"
        ),
    ] = [],
    exclude_tables: Annotated[
        List[str],
        typer.Option(help="List of tables that are excluded from the graph"),
    ] = [],
    include_tables: Annotated[
        List[str],
        typer.Option(help="List of tables that are included in the graph"),
    ] = [],
    python_dir: Annotated[
        List[Path],
        typer.Option(
            help="Paths to add to the `PYTHON_PATH` for module lookup.",
            file_okay=False,
            dir_okay=True,
            resolve_path=True,
            exists=True,
        ),
    ] = [],
    format: Annotated[
        Formats, typer.Option(help="The file format to output the generated graph to.")
    ] = Formats.mermaid,
):
    settings = get_pyproject_settings()
    base_class = get_base_class(base_class_path, settings)

    if settings and "imports" in settings:
        import_module.extend(settings["imports"])

    typer.echo(
        get_graph_string(
            base_class_path=base_class,
            import_module=import_module,
            include_tables=set(include_tables),
            exclude_tables=set(exclude_tables),
            python_dir=python_dir,
            format=format.value,
        )
    )


@app.command(help="Create a graph and inject it as a code field into a markdown file.")
def inject(
    file: Annotated[
        Path,
        typer.Argument(
            help="The file to inject the generated graph into.",
            file_okay=True,
            dir_okay=False,
            resolve_path=True,
            exists=True,
        ),
    ],
    base_class_path: Annotated[
        str,
        typer.Argument(help="The SQLAlchemy base class used by the database to graph."),
    ],
    replace_begin_tag: Annotated[
        str,
        typer.Option(help=""),
    ] = "<!-- BEGIN_SQLALCHEMY_DOCS -->",
    replace_end_tag: Annotated[
        str,
        typer.Option(help=""),
    ] = "<!-- END_SQLALCHEMY_DOCS -->",
    import_module: Annotated[
        List[str],
        typer.Option(
            help="Module, typically an SQL Model, to import. Modules that end in :* will act as `from module import *`"
        ),
    ] = [],
    exclude_tables: Annotated[
        List[str],
        typer.Option(help="List of tables that are excluded from the graph"),
    ] = [],
    include_tables: Annotated[
        List[str],
        typer.Option(help="List of tables that are included in the graph"),
    ] = [],
    python_dir: Annotated[
        List[Path],
        typer.Option(
            help="Paths to add to the `PYTHON_PATH` for module lookup.",
            file_okay=False,
            dir_okay=True,
            resolve_path=True,
            exists=True,
        ),
    ] = [],
    format: Annotated[
        Formats, typer.Option(help="The file format to output the generated graph to.")
    ] = Formats.mermaid,
    check: Annotated[
        bool,
        typer.Option(
            "--check",
            help="Perform a dry run and return a success code of 0 if there are no changes or 1 otherwise.",
        ),
    ] = False,
):
    # Generate Graph
    graph = get_graph_string(
        base_class_path=base_class_path,
        import_module=import_module,
        include_tables=set(include_tables),
        exclude_tables=set(exclude_tables),
        python_dir=python_dir,
        format=format.value,
    )

    comment_format = transformers[format].comment_format  # type: ignore

    # Convert Graph to Injection String
    graph_piece = f"""{replace_begin_tag}
```{comment_format}
{graph}
```
{replace_end_tag}"""

    # Get content from current file.
    with open(file, "r") as fp:
        old_content = fp.read()

    # Replace old content with newly generated content.
    pattern = re.escape(replace_begin_tag) + "(.*)" + re.escape(replace_end_tag)
    new_content = re.sub(pattern, graph_piece, old_content, flags=re.MULTILINE | re.DOTALL)

    # Return result depends on whether we're in check mode.
    if check:
        if new_content == old_content:
            # If content is the same then we passed the test.
            typer.echo("No changes detected.")
            sys.exit(0)
        else:
            # If content is different then we failed the test.
            typer.echo("Changes detected.")
            sys.exit(1)
    else:
        # Dump newly generated contents back to file.
        with open(file, "w") as fp:
            fp.write(new_content)


@app.command(help="Display the current installed version of paracelsus.")
def version():
    from . import _version

    typer.echo(_version.version)


if __name__ == "__main__":
    app()
